import cPickle
import math
import os
import random
import sys
import time

from pychart import theme
from pychart import canvas
from pychart import axis
from pychart import area
from pychart import line_plot
from pychart import legend
from pychart import text_box

from BTrees.IFBTree import IFSet
from zerodbext.catalog.indexes.field import fwscan_wins
from zerodbext.catalog.indexes.field import nbest_ascending_wins

theme.get_options()
theme.use_color = True
theme.scale_factor = 2

# db keys:
# 64
# 512
# 1024
# 2048
# 4096
# 8192
# 16384
# 32768
# 65536

class FieldIndexForwardSort:
    """ Benchmark and compare the field index forward sort algorithms """

    def __init__(self, limitbase=2, rlenbase=2, dbfn='sort.db', dbkey='65536'):
        self.limitbase = limitbase
        self.rlenbase = rlenbase
        self.dbfn = dbfn
        self.dbkey = dbkey
        self.index = self.get_index()
        self.numdocs = self.index._num_docs.value
        self.dbkey = dbkey
        # the set of rlens and limits are series generated via
        # exponents to the power of the base base, e.g.  [4, 16, 64,
        # 256, 1024, 4096, 16384, 65535 ] if numdocs = 65708 and the
        # base is 4
        self.rlens = series(self.numdocs+1, self.rlenbase)
        self.limits = series(self.numdocs+1, self.limitbase)

        self.sorts = (
            ('nbest', self.index.nbest_ascending),
            ('fwscan', self.index.scan_forward),
            ('timsort', self.index.timsort_ascending)
            )
 
    def get_index(self):
        if not os.path.exists(self.dbfn):
            raise NotImplementedError # XXX create index-creation code
        from ZODB.FileStorage.FileStorage import FileStorage
        from ZODB.DB import DB
        s = FileStorage(self.dbfn)
        db = DB(s, cache_size=300000)
        c = db.open()
        root = c.root()
        return root[self.dbkey]

    def __call__(self):
        if not os.path.exists('%s.pck' % self.dbkey):
            self.bench()
        self.chart()

    def bench(self):
        tf = open('%s.txt' % self.dbkey, 'w')

        def output(msg):
            tf.write(msg + '\n')
            tf.flush()
            print msg

        all_docids = list(self.index._rev_index.keys())
        random.shuffle(all_docids)

        main = []

        for rlen in self.rlens:
            docids = IFSet(random.sample(all_docids,
                                         min(rlen, len(all_docids))))
            output('for %s' % rlen)
            output('-----------------------------------')

            control = []
            for k, s in self.index._fwd_index.items():
                for docid in s:
                    if docid in docids:
                        control.append(docid)

            capture = {}
            result = None
            for name, fn in self.sorts:
                data = capture.setdefault(name, [])
                for limit in self.limits:
                    t, result = timer(fn, docids, limit)
                    result = list(result)
                    if control[:limit] != result:
                        raise AssertionError((control[:limit], result))
                    data.append(t)
                    output('%0.6f %s at limit %s' % (t, name, limit))

            main.append({'rlen':rlen, 'capture':capture})

        cPickle.dump(main, open('%s.pck' % self.dbkey, 'w'))

    def chart(self):

        self.main = cPickle.load(open('%s.pck' % self.dbkey))

        for chartable in self.main:
            self.detailchart(chartable)

        sortnames = [ x[0] for x in self.sorts ]

        for sortname1, sortname2 in product(sortnames, sortnames):
            if sortname1 == sortname2:
                continue
            self.comparisonchart(sortname1, sortname2)

    def detailchart(self, chartable):
        theme.reinitialize()

        min_y = 0
        max_y = 0
        capture = chartable.get('capture')
        for sortname, sortfn in self.sorts:
            data = capture[sortname]
            m = median(data)
            if m > max_y:
                max_y = m

        max_x = max(self.limits)
        min_x = min(self.limits)

        ipoints = 10.0

        x_interval = (max_x - min_x) / ipoints
        y_interval = (max_y - min_y) / ipoints

        xaxis = axis.X(label='Limit',
                       tic_interval = x_interval,
                       format='/4{}%d')
        yaxis = axis.Y(label='Seconds',
                       tic_interval = y_interval,
                       format='/4{}%0.3f')

        ar = area.T(
            x_range = (min_x, max_x),
            y_range = (min_y, max_y),
            x_axis  = xaxis,
            y_axis  = yaxis,
            legend =  legend.T(),
            )
        tb = text_box.T(loc=(140,90), text='Rlen\n%s' % chartable['rlen'])

        for sortname, sortfn in self.sorts:
            data = capture[sortname]
            linedata = [ (self.limits[x], data[x]) for x in range(len(data)) ]
            ar.add_plot(
                line_plot.T(label="%s" % sortname, data=linedata)
                )
        fd = open('detail-%s-%s.pdf' % (self.dbkey, chartable['rlen']), 'w')
        can = canvas.init(fd, 'pdf')
        ar.draw(can)
        tb.draw(can)
        can.close()

    def comparisonchart(self, sortname1, sortname2):
        linedata = []
        test_total = 0
        test_wrong = 0

        for rlendata in self.main:
            rlen = rlendata['rlen']
            capture = rlendata['capture']
            values1 = capture[sortname1]
            values2 = capture[sortname2]
            doc_ratio = rlen / float(self.numdocs)
            cutoff = None
            wins = []

            #test = sortname1 == 'fwscan' and sortname2 in ('nbest', 'timsort')
            #test_fn = fwscan_wins
            test = sortname1 == 'nbest' and sortname2 == 'timsort'
            test_fn = nbest_ascending_wins

            for x in xrange(0, min(len(values1), len(values2))):
                t1 = values1[x]
                t2 = values2[x]
                limit = self.limits[x]
                limitratio = limit / float(self.numdocs)
                won = t1 < t2
                if won:
                    wins.append(limit)

                wrongmsg = "wrong %s? rlen %s, limit %s (%0.5f > %0.5f)%s"

                if test:
                    test_total += 1
                    curvewin = test_fn(limit, rlen, self.numdocs)
                    if won and (not curvewin):
                        extra = ''
                        if (t1 /  t2) < .90: # more than 10% difference
                            extra = " * (%0.2f)" % (t1/t2)
                        print wrongmsg % ('curvelose', rlen, limit, t2, t1,
                                          extra)
                        test_wrong +=1
                    elif (not won) and curvewin:
                        extra = ''
                        if (t2 /  t1) < .90: # more than 10% difference
                            extra = " * (%0.2f)" % (t2/t1)
                        print wrongmsg % ('curvewin', rlen, limit, t1, t2,
                                          extra)
                        test_wrong +=1

            for limit in wins:
                limitratio = limit / float(self.numdocs)
                linedata.append((doc_ratio, limitratio))

        if test:
            if test_total:
                test_right = test_total - test_wrong
                test_percent = test_right / float(test_total)
                print "test percentage %0.2f: (%s wrong out of %s)" % (
                    test_percent, test_wrong, test_total)

        comparename = 'compare-%s-%s-beats-%s' % (self.dbkey,
                                                  sortname1, sortname2)

        xaxis=axis.X(label='Doc Ratio (rlen//numdocs)',
                     tic_interval=.1,
                     format='/4{}%0.2f')
        yaxis=axis.Y(label='Limit Ratio (limit//numdocs)',
                     tic_interval=.1,
                     format='/4{}%0.2f')

        ar = area.T(
            x_range = (0, 1),
            y_range = (0, 1),
            x_axis  = xaxis,
            y_axis  = yaxis,
            legend =  legend.T(),
            )

        ar.add_plot(
            line_plot.T(label="%s \nbeats \n%s" % (sortname1, sortname2),
                        data=linedata),
            )

        tb = text_box.T(loc=(140,90), text='Numdocs\n%s' % self.numdocs)

        fd = open('%s.pdf' % comparename, 'w')
        can = canvas.init(fd, 'pdf')
        ar.draw(can)
        tb.draw(can)
        can.close()

def timer(fn, *args, **kw):
    times = []
    for x in xrange(7):
        start = time.time()
        result = fn(*args, **kw)
        if not hasattr(result, '__len__'):
            result = list(result)
        end = time.time()
        times.append(end-start)
    return median(times), result

def isect(seq1, seq2):
    res = []                     # start empty
    for x in seq1:               # scan seq1
        if x in seq2:            # common item?
            res.append(x)        # add to end
    return res

def product(*args):
    # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
    # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
    pools = map(tuple, args)
    result = [[]]
    for pool in pools:
        result = [x+[y] for x in result for y in pool]
    for prod in result:
        yield tuple(prod)

def avg(numbers):
    return sum(numbers) / len(numbers)

def median(numbers):
    "Return the median of the list of numbers."
    # Sort the list and take the middle element.
    n = len(numbers)
    copy = numbers[:] # So that "numbers" keeps its original order
    copy.sort()
    if n & 1:         # There is an odd number of elements
        return copy[n // 2]
    else:
        return (copy[n // 2 - 1] + copy[n // 2]) / 2
  
def series(numdocs, base):
    exp = int(math.ceil(math.log(numdocs) / math.log(base)))
    return [ pow(base, x) for x in range(1, exp) ]

def main(argv=sys.argv):
    bench = FieldIndexForwardSort()
    bench()





class RangeValue:
    """ Use in fieldindex query above to indicate a range search for a term """
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def as_tuple(self):
        return (self.start, self.end)

Range = RangeValue # deprecated BBB



import six
from persistent import Persistent
from ZODB.broken import Broken
import BTrees

_marker = ()


class CatalogIndex(object):
    """ Abstract class for interface-based lookup """

    family = BTrees.family32

    def __init__(self, discriminator):
        if not callable(discriminator):
            if not isinstance(discriminator, six.string_types):
                raise ValueError('discriminator value must be callable or a '
                                 'string')
        self.discriminator = discriminator
        self._not_indexed = self.family.IF.Set()

    def index_doc(self, docid, object):
        if callable(self.discriminator):
            value = self.discriminator(object, _marker)
        else:
            value = getattr(object, self.discriminator, _marker)

        if value is _marker:
            # unindex the previous value
            super(CatalogIndex, self).unindex_doc(docid)

            # Store docid in set of unindexed docids
            self._not_indexed.add(docid)

            return None

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object %s' %
                             value)

        if isinstance(value, Broken):
            raise ValueError('Catalog cannot index broken object %s' %
                             value)

        if docid in self._not_indexed:
            # Remove from set of unindexed docs if it was in there.
            self._not_indexed.remove(docid)

        return super(CatalogIndex, self).index_doc(docid, value)

    def unindex_doc(self, docid):
        _not_indexed = self._not_indexed
        if docid in _not_indexed:
            _not_indexed.remove(docid)
        super(CatalogIndex, self).unindex_doc(docid)

    def reindex_doc(self, docid, object):
        """ Default reindex_doc implementation """
        self.unindex_doc(docid)
        self.index_doc(docid, object)

    def docids(self):
        not_indexed = self._not_indexed
        indexed = self._indexed()
        if len(not_indexed) == 0:
            return self.family.IF.Set(indexed)
        elif len(indexed) == 0:
            return not_indexed
        indexed = self.family.IF.Set(indexed)
        return self.family.IF.union(not_indexed, indexed)

    def apply_intersect(self, query, docids):
        """ Default apply_intersect implementation """
        result = self.apply(query)
        if docids is None:
            return result
        return self.family.IF.weightedIntersection(result, docids)[1]

    def _negate(self, assertion, *args, **kw):
        positive = assertion(*args, **kw)
        all = self.docids()
        if len(positive) == 0:
            return all
        return self.family.IF.difference(all, positive)

    def applyContains(self, *args, **kw):
        raise NotImplementedError(
            "Contains is not supported for %s" % type(self).__name__)

    def applyDoesNotContain(self, *args, **kw):
        return self._negate(self.applyContains, *args, **kw)

    def applyEq(self, *args, **kw):
        raise NotImplementedError(
            "Eq is not supported for %s" % type(self).__name__)

    def applyNotEq(self, *args, **kw):
        return self._negate(self.applyEq, *args, **kw)

    def applyGt(self, *args, **kw):
        raise NotImplementedError(
            "Gt is not supported for %s" % type(self).__name__)

    def applyLt(self, *args, **kw):
        raise NotImplementedError(
            "Lt is not supported for %s" % type(self).__name__)

    def applyGe(self, *args, **kw):
        raise NotImplementedError(
            "Ge is not supported for %s" % type(self).__name__)

    def applyLe(self, *args, **kw):
        raise NotImplementedError(
            "Le is not supported for %s" % type(self).__name__)

    def applyAny(self, *args, **kw):
        raise NotImplementedError(
            "Any is not supported for %s" % type(self).__name__)

    def applyNotAny(self, *args, **kw):
        return self._negate(self.applyAny, *args, **kw)

    def applyAll(self, *args, **kw):
        raise NotImplementedError(
            "All is not supported for %s" % type(self).__name__)

    def applyNotAll(self, *args, **kw):
        return self._negate(self.applyAll, *args, **kw)

    def applyInRange(self, *args, **kw):
        raise NotImplementedError(
            "InRange is not supported for %s" % type(self).__name__)

    def applyNotInRange(self, *args, **kw):
        return self._negate(self.applyInRange, *args, **kw)

    def _migrate_to_0_8_0(self, docids):
        """
        I'm sorry.
        """
        docids = self.family.IF.Set(docids)
        indexed = self.family.IF.Set(self._indexed())
        self._not_indexed = self.family.IF.difference(docids, indexed)



import unittest

FACETS = [
    'price',
    'price:0-100',
    'price:100-500',
    'price:100-*',
    'color',
    'color:blue',
    'color:red',
    'size',
    'size:small',
    'size:medium',
    'size:large',
    'style',
    'style:gucci',
    'style:gucci:handbag',
    'style:gucci:dress',
    ]

_marker = object()

class TestCatalogFacetIndex(unittest.TestCase):
    def _getTargetClass(self):
        from zerodbext.catalog.indexes.facet import CatalogFacetIndex
        return CatalogFacetIndex

    def _makeOne(self, discriminator=None, facets=FACETS, family=_marker):
        def _discriminator(obj, default):
            return obj
        if discriminator is None:
            discriminator = _discriminator
        if family is _marker:
            return self._getTargetClass()(discriminator, facets)
        return self._getTargetClass()(discriminator, facets, family)

    def _populateIndex(self, idx):
        idx.index_doc(1, ['price:0-100', 'color:blue', 'style:gucci:handbag'])
        idx.index_doc(2, ['price:0-100', 'color:blue', 'style:gucci:dress'])
        idx.index_doc(3, ['price:0-100', 'color:red', 'color:blue',
                          'style:gucci'])
        idx.index_doc(4, ['size:large'])

    def test_class_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyClass
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyClass(ICatalogIndex, self._getTargetClass())

    def test_instance_conforms_to_ICatalogIndex(self):
        from zope.interface.verify import verifyObject
        from zerodbext.catalog.interfaces import ICatalogIndex
        verifyObject(ICatalogIndex, self._makeOne())

    def test_ctor_defaults(self):
        from BTrees import family32
        index = self._makeOne()
        self.assertTrue(index.discriminator(self, index) is self)
        self.assertEqual(list(index.facets), sorted(FACETS))
        self.assertTrue(index.family is family32)

    def test_ctor_explicit(self):
        from BTrees import family64
        OTHER_FACETS = ['foo', 'foo:bar']
        def _discriminator(obj, default):
            return default
        index = self._makeOne(_discriminator, OTHER_FACETS, family64)
        self.assertTrue(index.discriminator(self, index) is index)
        self.assertEqual(list(index.facets), OTHER_FACETS)
        self.assertTrue(index.family is family64)

    def test_ctor_string_discriminator(self):
        from BTrees import family64
        OTHER_FACETS = ['foo', 'foo:bar']
        index = self._makeOne('facets')
        self.assertEqual(index.discriminator, 'facets')

    def test_ctor_bad_discriminator(self):
        self.assertRaises(ValueError, self._makeOne, object())

    def test_index_doc_callback_discriminator(self):
        OTHER_FACETS = ['foo', 'foo:bar', 'foo:baz']
        def _discrimintator(obj, default):
            return ['foo:bar']
        index = self._makeOne(_discrimintator, OTHER_FACETS)
        index.index_doc(1, object())
        self.assertEqual(list(index._fwd_index['foo']), [1])
        self.assertEqual(list(index._fwd_index['foo:bar']), [1])
        self.assertEqual(list(index._rev_index[1]), ['foo', 'foo:bar'])

    def test_index_doc_string_discriminator(self):
        OTHER_FACETS = ['foo', 'foo:bar', 'foo:baz']
        class Dummy:
            facets = ['foo:bar']
        index = self._makeOne('facets', OTHER_FACETS)
        index.index_doc(1, Dummy())
        self.assertEqual(list(index._fwd_index['foo']), [1])
        self.assertEqual(list(index._fwd_index['foo:bar']), [1])
        self.assertEqual(list(index._rev_index[1]), ['foo', 'foo:bar'])

    def test_index_doc_missing_value_unindexes(self):
        OTHER_FACETS = ['foo', 'foo:bar', 'foo:baz']
        class Dummy:
            pass
        dummy = Dummy()
        dummy.facets = ['foo:bar']
        index = self._makeOne('facets', OTHER_FACETS)
        index.index_doc(1, dummy)
        del dummy.facets
        index.index_doc(1, dummy)
        self.assertFalse('foo' in index._fwd_index)
        self.assertFalse('foo:bar' in index._fwd_index)
        self.assertFalse(1 in index._rev_index)

    def test_index_doc_persistent_value_raises(self):
        from persistent import Persistent
        OTHER_FACETS = ['foo', 'foo:bar', 'foo:baz']
        class Dummy:
            pass
        index = self._makeOne('facets', OTHER_FACETS)
        dummy = Dummy()
        dummy.facets = Persistent()
        self.assertRaises(ValueError, index.index_doc, 1, dummy)

    def test_index_doc_unindexes_old_values(self):
        OTHER_FACETS = ['foo', 'foo:bar', 'foo:baz']
        class Dummy:
            pass
        dummy = Dummy()
        dummy.facets = ['foo:bar']
        index = self._makeOne('facets', OTHER_FACETS)
        index.index_doc(1, dummy)
        dummy.facets = ['foo:baz']
        index.index_doc(1, dummy)
        self.assertEqual(list(index._fwd_index['foo']), [1])
        self.assertEqual(list(index._fwd_index['foo:baz']), [1])
        self.assertEqual(list(index._rev_index[1]), ['foo', 'foo:baz'])
        self.assertFalse('foo:bar' in index._fwd_index)

    def test_search(self):
        index = self._makeOne()
        self._populateIndex(index)

        result = index.search(['color:blue', 'color:red'])
        self.assertEqual(sorted(list(result)), [3])
        result = index.search(['price'])
        self.assertEqual(sorted(list(result)), [1,2,3])
        result = index.search(['price:0-100'])
        self.assertEqual(sorted(list(result)), [1,2,3])
        result = index.search(['style:gucci'])
        self.assertEqual(sorted(list(result)), [1,2,3])
        result = index.search(['style:gucci:handbag'])
        self.assertEqual(sorted(list(result)), [1])
        result = index.search(['size'])
        self.assertEqual(sorted(list(result)), [4])
        result = index.search(['size:large'])
        self.assertEqual(sorted(list(result)), [4])
        result = index.search(['size:nonexistent'])
        self.assertEqual(sorted(list(result)), [])
        result = index.search(['nonexistent'])
        self.assertEqual(sorted(list(result)), [])

        index.unindex_doc(1)
        result = index.search(['price'])
        self.assertEqual(sorted(list(result)), [2,3])
        result = index.search(['price:0-100'])
        self.assertEqual(sorted(list(result)), [2,3])
        result = index.search(['style:gucci'])
        self.assertEqual(sorted(list(result)), [2,3])
        result = index.search(['style:gucci:handbag'])
        self.assertEqual(sorted(list(result)), [])

        index.unindex_doc(2)
        result = index.search(['price'])
        self.assertEqual(sorted(list(result)), [3])
        result = index.search(['price:0-100'])
        self.assertEqual(sorted(list(result)), [3])
        result = index.search(['style:gucci'])
        self.assertEqual(sorted(list(result)), [3])

        index.unindex_doc(4)
        result = index.search(['size'])
        self.assertEqual(sorted(list(result)), [])
        result = index.search(['size:large'])
        self.assertEqual(sorted(list(result)), [])
        result = index.search(['size:nonexistent'])
        self.assertEqual(sorted(list(result)), [])
        result = index.search(['nonexistent'])
        self.assertEqual(sorted(list(result)), [])

    def test_counts(self):
        index = self._makeOne()
        self._populateIndex(index)

        search = ['price:0-100']
        result = index.search(search)
        self.assertEqual(sorted(list(result)), [1,2,3])
        counts = index.counts(result, search)
        self.assertEqual(counts['style'], 3)
        self.assertEqual(counts['style:gucci'], 3)
        self.assertEqual(counts['style:gucci:handbag'], 1)
        self.assertEqual(counts['style:gucci:dress'], 1)
        self.assertEqual(counts['color'], 3)
        self.assertEqual(counts['color:blue'], 3)
        self.assertEqual(counts['color:red'], 1)
        self.assertEqual(len(counts), 7)

        search = ['price:0-100', 'color:red']
        result = index.search(search)
        self.assertEqual(sorted(list(result)), [3])
        counts = index.counts(result, search)
        self.assertEqual(counts['style'], 1)
        self.assertEqual(counts['style:gucci'], 1)
        self.assertEqual(counts['color:blue'], 1)
        self.assertEqual(len(counts), 3)

        search = ['size:large']
        result = index.search(search)
        self.assertEqual(sorted(list(result)), [4])
        counts = index.counts(result, search)
        self.assertEqual(counts, {})

        search = ['size']
        result = index.search(search)
        self.assertEqual(sorted(list(result)), [4])
        counts = index.counts(result, search)
        self.assertEqual(counts, {'size:large':1})

    def test__indexed(self):
        index = self._makeOne()
        self._populateIndex(index)
        self.assertEqual(set(index._indexed()), set((1, 2, 3, 4)))

    def test_index_doc_missing_value_adds_to__not_indexed(self):
        def discriminator(obj, default):
            return default
        index = self._makeOne(discriminator)
        self.assertEqual(index.index_doc(20, 3), None)
        self.assertTrue(20 in index._not_indexed)

    def test_index_doc_with_value_removes_from__not_indexed(self):
        index = self._makeOne()
        index._not_indexed.add(20)
        self.assertEqual(index.index_doc(20, 'foo'), 'foo')
        self.assertFalse(20 in index._not_indexed)

